#ifndef __3D
#define __3D
#define PI 3.1415926535897932385

#define sin256(n) sinTable[(unsigned char)(n)]
#define cos256(n) cosTable[(unsigned char)(n)]

extern float sinTable[256];		//precalculated sin and cos tables
extern float cosTable[256];


struct TVector
{
	float x,y,z;
};

struct TQuat		//a quaternion representation
{
	float x,y,z,w;
};

struct TPolygon
{
	int* vertices;	//array of indexes to vertices
	int vertCount;	//no. of vertices
	int centre, normal;	//indexes of the approximate centre and normal vertices
	unsigned long colour; //24bit RGB
	char side;			//0=both sides, 1 , -1 just one side (difficult to tell which one :-) )
};

struct TCam			//a camera structure
{
	TVector pos;	//position
	TQuat quat;		//rotation

	float dist;		//eye-screen distance
	float shiftRight;	//half of the 3D cha-cha distance (negative for one of the pictures)
};


void rotateVectors(TVector* v, TQuat* q, int count=1);	//rotate count vectors from the array v by the quat *q
void quatMult(TQuat* dest, TQuat* p, TQuat* q);		//compose two rotations dest=p*q
void precalcSinCos();		//fill the precalculated tables
#endif

