#ifndef __OBJECT
#define __OBJECT
#include "screen.h"
#include "3d.h"

#include <netinet/in.h>

typedef uint32_t int32;

struct BSP_Vector		//used when building the object
{
	TVector v;
	int index;
	BSP_Vector* next;
};

struct BSP_Poly  		//used when building the object
{
	float d;
	int vertCount;
	unsigned long colour;



	char side;
	BSP_Vector** vertices;
	BSP_Vector* normal;
	BSP_Poly* next;
};

struct BSP_Tree  	//a BSP tree node
{
	union {
			TPolygon *poly;		//used most of the time
			BSP_Poly *bsppoly;	//used when building the object
		};
	BSP_Tree *pos, *neg;		//positive and negative branches
};


class TObject			//main class for a viewable 3d object composed of polygons
{
public:
	TObject(TObject *);		//constructor for non-templates (ordinary instances)
	TObject(char *, float mag=1);		//template constructor

	virtual ~TObject();
	void draw(TScreen* scr, TCam* cam, TVector* lght);

	inline char isOK() {return ok;}			//1 if ok, 0=serious problem
	inline TVector getPos() {return pos;}
	inline void setPos(TVector v) {pos=v;}
	inline float getRadius() {return radius;}

	inline TQuat getQuat() {return quat;}
	inline void setQuat(TQuat* q) {quat=*q;}
	inline unsigned long getAvgColour() {return avgColour;}

	virtual char* dump(char *buff);	//dumps all the instance-specific data into buff for network-cloning
	virtual char* read(char *buff);  //reads all the instance-specific data from buff

protected:
	TVector pos;			//position of the object's centre
	TQuat quat;				//rotation
	char buildBSPTree(BSP_Tree** root, BSP_Poly** polyList, BSP_Vector** vertList);	//used when building the template
	void drawPolygons(BSP_Tree* t, TScreen* scr, TCam* cam, TVector* light);  //called recursively from draw(...) - walks the BSP tree

	char ok;

	BSP_Tree* bspt;	  //(shared by instances)

	TObject* original;		//0 if this is the template, otherwise pointer to the template

	float radius;			//approximate size of the object
	unsigned long avgColour;	//approx. average colour

	int polyCount;			//no. of polygons

	TVector* vertices;		//vertices array
	int vertCount,normalCount;	//no. of vertices, no. of normals among them (they are handled specially, but are in this array to save time when rotating)

	int* vertices2D;	//a temporary array for storing on-screen coordinates while drawing (shared by instances)
};


//macros used for packing and restoring floats into/from network-order ints
#define PACK(n, precision) htonl( (long)((n)*(precision)) )
#define UNPACK(n, precision) ( ((float)((signed long)ntohl(n))) /(precision))


#endif

