#include "screen.h"
#include <alloc.h>
#include <math.h>
#include <stdlib.h>
#include <gdk/gdk.h>
#include "const.h"
#include <stdio.h>
TScreen::TScreen(void* p)
{
	gdk_rgb_init();
	win = (GdkWindow*)p;
	pix=0;
	wingc = gdk_gc_new(win);
	ok=1;
	
}

TScreen::~TScreen()
{
	gdk_gc_unref(wingc);
	free_pixmap();	

}

void TScreen::configure()
{	
	static int oldw=-2, oldh=-2;
	gint w,h;
	
	gdk_window_get_size(win, &w, &h);
	scrWidth = w;
	scrHeight = h;
	if (oldw!=w || oldh!=h)
	{
		free_pixmap();
		alloc_pixmap();
		oldw=w; oldh=h;
	}
}

void TScreen::alloc_pixmap()
{
	pix=gdk_pixmap_new(win, scrWidth, scrHeight, -1);
	pixgc=gdk_gc_new(pix);
}

void TScreen::free_pixmap()
{
	if (pix) gdk_pixmap_unref(pix);
	if (pixgc) gdk_gc_unref(pixgc);
}
	

void TScreen::clipRect(int x1, int y1, int x2, int y2)
	//all the TScreen methods support clipping and all the coordinates
	//are given relative to the current clip rectangle
	//cliprect can be used to change the rectangle
{
	//adjust left, top, width and height so they are all on-screen
	//this is being relied on in other methods!!!
	left=x1; top=y1;
	if (left<0) left=0;
	if (top<0) top=0;

	if (x2<0) x2=0;
	if (y2<0) y2=0;

	if (x2>=scrWidth) x2=scrWidth;
	if (y2>=scrHeight) y2=scrHeight;

	if (left>x2) left=x2;
	if (top>y2) top=y2;

	width=x2-left;
	height=y2-top;
	GdkRectangle r = {left, top, width, height};
	gdk_gc_set_clip_rectangle(pixgc, &r);
}

//draw a pixel with the colour c. If c == -1, the pen colour is used
void TScreen::putPixel(int x, int y, long c)
{
	if (c>0) setcolor(c);
	gdk_draw_point(pix, pixgc, x+left, y+top); 	
	if (c>0) setcolor(pen);
}

//copies the off-screen buffer to the video RAM
void TScreen::redraw()
{
	gdk_draw_pixmap(win, wingc, pix, 0, 0, 0, 0, scrWidth, scrHeight);
}


//sets all the pixels to the color given by c
void TScreen::clrscr(unsigned long c)
{
	setcolor(c);
	gdk_draw_rectangle(pix , pixgc, TRUE, left, top, width, height);
	setcolor(pen);
}

//draws a line from [x1,y1] to [x2,y2]
void TScreen::line(int _x1, int _y1, int _x2, int _y2)
{
	gdk_draw_line(pix, pixgc, _x1+left, _y1+top, _x2+left, _y2+top);
}

/* polygon(...) draws a (count)-vertex polygon
(points) is an array of coordinates: x1,y1,x2,y2,x3,y3....
(indexes) is an array of indexes into the (points) array, specifying
the order of the points in the polygon
*/

void TScreen::polygon(int count, int *points, int* indexes)
{
	//printf("poly!\n");
	int i;
	GdkPoint p[MAX_POLY_POINT_CNT];
	
	for (i=0; i<count; ++i)
	{
		p[i].x=left+points[2*indexes[i]];
		p[i].y=top+points[2*indexes[i]+1];
	}
	gdk_draw_polygon(pix, pixgc, TRUE, p, count); 
}

//rectangle(...) draws a rectangle, parallel with the screen edges
void TScreen::rectangle(int x1, int y1, int x2, int y2)
{
	int t;
	if (x1>x2) {t=x1; x1=x2; x2=t;}
	if (y1>y2) {t=y1; y1=y2; y2=t;}
	gdk_draw_rectangle(pix, pixgc, TRUE, x1+left, y1+top, x2-x1, y2-y1);
	
}

void TScreen::setcolor(unsigned long c)
{
	gdk_rgb_gc_set_foreground(pixgc, c);
	gdk_rgb_gc_set_background(pixgc, c);
}

/********************* TAnaglyphScreen ************************/
/*
TAnaglyphScreen supports the anaglyph 3d-viewing technique.
The main difference from TScreen is that it has two
off-screen buffers, one for the left eye and one for the right eye.
You can switch between the eyes using selectEye(eye).

Draw the scene twice, once for each eye, shifting the camera between the
two shots. redraw() then creates an anaglyph image.
*/
TAnaglyphScreen::TAnaglyphScreen(void* p):TScreen(p)
{
	pixL=pixR=0;
}

TAnaglyphScreen::~TAnaglyphScreen()
{

}


//selects the appropriate off-screen buffer
void TAnaglyphScreen::selectEye(char n)
{
	if (!stereo) n=1;
	eye=n;
	pix=eye?pixR:pixL;
}

void TAnaglyphScreen::redraw()
{
	
	clipRect(0,0, scrWidth, scrHeight);
	if (stereo) {
		gdk_gc_set_function(pixgc, GDK_OR);
		gdk_draw_pixmap(pixR, pixgc, pixL, 0, 0, 0, 0, scrWidth, scrHeight);
	}
	gdk_gc_set_function(pixgc, GDK_COPY);
	gdk_draw_pixmap(win, wingc, pixR, 0, 0, 0, 0, scrWidth, scrHeight);

}

void TAnaglyphScreen::alloc_pixmap()
{
	pixL=gdk_pixmap_new(win, scrWidth, scrHeight, -1);
	pixR=gdk_pixmap_new(win, scrWidth, scrHeight, -1);
	pixgc=gdk_gc_new(pixL);
}

void TAnaglyphScreen::free_pixmap()
{
	if (pixL) gdk_pixmap_unref(pixL);
	if (pixR) gdk_pixmap_unref(pixR);
	if (pixgc) gdk_gc_unref(pixgc);
}

void TAnaglyphScreen::setcolor(unsigned long c)	
{
	if (stereo)
		if (eye) c&=clRed;
			else c&=clCyan;
	TScreen::setcolor(c);
}	
