#ifndef __SCREEN
#define __SCREEN

#include <gdk/gdk.h>

#define clRed 0xFF0000
#define clGreen 0x00FF00
#define clBlue 0x0000FF

#define clYellow (clRed | clGreen)
#define clCyan (clBlue | clGreen)
#define clMagenta (clBlue | clRed)

#define clWhite (clRed | clGreen | clBlue)

#define shade(c,br) ( (((((unsigned long)(c)&clRed) * (br)) >> 8)&clRed) |		\
					(((((unsigned long)(c)&clGreen) * (br)) >> 8)&clGreen)	|\
					(((((unsigned long)(c)&clBlue) * (br)) >> 8)&clBlue)		)
	//shade(c, n) a shade of a given colour gives c for n==255, clBlack for n==0


//TScreen supports lines, rectangles and pixels. It draws off-screen until
//told to redraw the hardware screen. Supports clipping.
//All coordinates are relative to the current clipping rectangle
//Curently uses VESA 320*200*16bit
class TScreen
{

public:
	TScreen(void *);
	virtual ~TScreen();

	
	void putPixel(int x, int y, long c);

	void clrscr(unsigned long c=0);	//clears the current clipping rectangle, default background is black
	void line(int x1,int y1,int x2,int y2);	//line from [x1,y1] to [x2,y2]
	void TScreen::polygon(int count, int *points, int *indexes);	//draws an "count-agon". Accepts an array of (x,y) integer pairs and an array of indices to the first array
	void TScreen::rectangle(int x1, int y1, int x2, int y2);	//a rectangle where [x1,y1] and [x2, y2] are opposite corners
	void redraw();			//draws the contents of th off-screen buffer on the screen
	void clipRect(int x1, int y1, int x2, int y2);	//sets the clipping rectangle
	inline int getheight() {return height;};	//clipping height
	inline int getwidth() {return width;};		//clipping width
	inline int getscrheight() {return scrHeight;};	//full-screen height
	inline int getscrwidth() {return scrWidth;};	//full-screen width
	char isOK() {return ok;}		//1 = ok, 0=big problem
	void setpen(unsigned long c) {pen=c; setcolor(pen);};
	void configure();
	char stereo;


protected:
	char ok;
	int left, top, width, height;
	int scrWidth, scrHeight;
	GdkGC *wingc, *pixgc;
	
	GdkPixmap* pix;
	GdkWindow* win;
	virtual void setcolor(unsigned long c);	
	unsigned long pen;		//active colour
	virtual void alloc_pixmap();
	virtual void free_pixmap();
	

};


//TAnaglyph screen extends TScreen by adding a second off-screen buffer.
//An image is drawn to one off the buffers, then the camera is shifted and the image
//is drawn to the other one. Redraw then composes these two images to
//create an anaglyph image for viewing with red-cyan stereo anaglyph glasses.

class TAnaglyphScreen:public TScreen
{
public:
	TAnaglyphScreen(void* p);
	virtual ~TAnaglyphScreen();
	void redraw();
	void selectEye(char n);
	GdkPixmap *pixR, *pixL;
	virtual void setcolor(unsigned long c);	
	char stereo;

private:
	virtual void alloc_pixmap();
	virtual void free_pixmap();
	char eye;
	
};


#endif

