#include "screen.h"
#include<mem.h>
#include<alloc.h>
#include<math.h>
#include<stdlib.h>


TScreen::TScreen()
{
	snowCheck=1;
	pen=1;
	scrWidth=320;
	scrHeight=200;

	(long)scr=0xa0000000L;	//viedo memory
	(void far*)buff=farmalloc(64000);	//allocate the off-screen buffer

	clipRect(0,0, 9999, 9999);	//clipRect sets the correct left, top, width and height for clipping

	ok=buff?1:0;	//success flag
}

TScreen::~TScreen()
{
	if (buff) farfree(buff);
}


void TScreen::clipRect(int x1, int y1, int x2, int y2)
	//all the TScreen methods support clipping and all the coordinates
	//are given relative to the current clip rectangle
	//cliprect can be used to change the rectangle
{
	//adjust left, top, width and height so they are all on-screen
	//this is being relied on in other methods!!!
	left=x1; top=y1;
	if (left<0) left=0;
	if (top<0) top=0;

	if (x2<0) x2=0;
	if (y2<0) y2=0;

	if (x2>=scrWidth) x2=scrWidth;
	if (y2>=scrHeight) y2=scrHeight;

	if (left>x2) left=x2;
	if (top>y2) top=y2;

	width=x2-left;
	height=y2-top;
}

void TScreen::putPixel(int x, int y, unsigned char c)
{
	if (x>=0 && x<width && y>=0 && y<height)
	buff[top+y][left+x]=c;
}

unsigned char TScreen::getPixel(int x, int y)
{
	return(buff[top+y][left+x]);
}

//Wait-for-vertical-retrace is used before a screen redraw to
//avoid flicker
void TScreen::wfvr()
{
	if (snowCheck)
	{
			asm MOV  DX, 0x3DA;

		RAS1:
			asm IN   AL, DX;
			asm TEST AL, 8;
			asm JNZ  RAS1;

		RAS2:
			asm IN   AL, DX;
			asm TEST AL, 8;
			asm JZ   RAS2;
	}
}

//copies the off-screen buffer to the video RAM
void TScreen::redraw()
{
	wfvr();
	_fmemcpy(scr,buff,64000);
}

//swithes between 320*200*256 and the original mode
//in the graphics mode, sets the pallete to 2+2+2bit RGB
void TScreen::graphicsMode(char onOff)
{
	//swith the mode
	if (onOff)
	{
		asm {
			mov ah, 0xf
			int 0x10
		}
		oldMode=_AL;
		_AL=0x13;

	} else _AL=oldMode;
	asm {
		mov ah,0
		int 0x10
	}

	//graphics? If so, change the pallette
	if  (onOff)
	{
		for (unsigned char i=0; i<64;++i)
		{
			char r,g,b;

			r=16*(i&3);
			g=4*(i&12);
			b=(i&48);

			asm{
				mov dx,0x3c6
				mov al, 0xff
				out dx,al

				mov al,i
				mov dx,0x3c8
				out dx,al

				mov dx,0x3c9

				mov al,r
				out dx,al

				mov al,g
				out dx,al

				mov al,b
				out dx,al
			}
		}
	}
}

//sets all the pixels to the color given by c
void TScreen::clrscr(unsigned char c)
{
	for (int i=0; i<height; ++i)
		_fmemset(buff[top+i]+left, c, width);
}

//draws a line from [x1,y1] to [x2,y2]
void TScreen::line(int _x1, int _y1, int _x2, int _y2)
{
	long x1,y1,x2,y2;	//against overflow!
	x1=_x1; y1=_y1;	x2=_x2; y2=_y2;

	//clip off the line
	int tmp;
	if (x1>x2) 	//make sure x1<=x2
	{
		tmp=x1; x1=x2; x2=tmp;
		tmp=y1; y1=y2; y2=tmp;
	}

	if (x1>=width || x2<0) return;	//nothing to draw


	if (x1<0)	//left clip
	{
		y1=y1+(y2-y1)*(0-x1)/(x2-x1);
		x1=0;
	}

	if (x2>=width)	//right clip
	{
		y2=y1+(y2-y1)*(width-x1)/(x2-x1);
		x2=width-1;
	}


	if (y1>y2) 	//make sure y1<=y2
	{
		tmp=x1; x1=x2; x2=tmp;
		tmp=y1; y1=y2; y2=tmp;
	}

	if (y1>=height || y2<0) return;	//nothing to draw


	if (y1<0)	//top clip
	{
		x1=x1+(x2-x1)*(0-y1)/(y2-y1);
		y1=0;
	}

	if (y2>=height)	//bottom clip
	{
		x2=x1+(x2-x1)*(height-y1)/(y2-y1);
		y2=height-1;
	}


	// a fast graph algorithm...
	int dx=abs(x2-x1);
	int dy=abs(y1-y2);


	int incA, incB;

	if (dx>dy)
	{
		incA=x1>x2?-1:1;
		incB=y1>y2?-320:320;
	} else
	{

		tmp=dx; dx=dy; dy=tmp;

		incA=y1>y2?-320:320;
		incB=x1>x2?-1:1;
	}

	unsigned char* p=*buff+x1+left+320*(y1+top);

	long trigB=dx/2;
	long trigAt=dx;
	for (int i=dx;i>0;--i)
	{
		*p=pen;
		p+=incA;

		trigB+=dy;
		if (trigB>=trigAt)
		{
			p+=incB;
			trigAt+=dx;
		}
	}
}

/* polygon(...) draws a (count)-vertex polygon
(points) is an array of coordinates: x1,y1,x2,y2,x3,y3....
(indexes) is an array of indexes into the (points) array, specifying
the order of the points in the polygon
*/

void TScreen::polygon(int count, int *points, int* indexes)
{

	int left[200], right[200];
	int min=200, max=0; //min and max y coordinate
	for (int i=0; i<count; ++i)
	{
		int y=points[2*indexes[i]+1];
		if (min>y) min=y;
		if (max<y) max=y;
	}
	if (min<0) min=0;
	if (max>=200) max=200-1;

	for (i=min; i<=max; ++i)
	{
		left[i]=320;
		right[i]=-1;
	}

	int x2=points[2*indexes[0]];
	int y2=points[2*indexes[0]+1];


	for (int j=1; j<count+1; ++j)
	{
		int x1=x2;
		int y1=y2;

		int index=(j<count)?j:0;
		x2=points[2*indexes[index]];
		y2=points[2*indexes[index]+1];

		int dx=abs(x1-x2);
		int dy=abs(y1-y2);


		int incAX, incAY, incBX, incBY;

		if (dx>dy)
		{
			incAX=x1>x2?-1:1;
			incAY=0;
			incBX=0;
			incBY=y1>y2?-1:1;

		} else
		{

			int tmp=dx; dx=dy; dy=tmp;

			incAX=0;
			incAY=y1>y2?-1:1;
			incBX=x1>x2?-1:1;
			incBY=0;
		}

		int x=x1;
		int y=y1;

		long trigB=dx/2;
		long trigAt=dx;
		for (int i=dx;i>0;--i)
		{

			if (y>=0 && y<200)
			{
				if (left[y]>x) left[y]=x;
				if (right[y]<x) right[y]=x;
			}

			x+=incAX;
			y+=incAY;

			trigB+=dy;
			if (trigB>=trigAt)
			{
				x+=incBX;
				y+=incBY;

				trigAt+=dx;
			}
		}


	}
	// The left-right arrays are filled up. Now, draw the polygon;
	unsigned char* p0=(unsigned char*)buff[min+top]+this->left;
	if (max>=height) max=height-1;


	for (i=min; i<=max; ++i)
	{
		if (left[i]<0) left[i]=0;
		if (right[i]>=width) right[i]=width-1;

		if (right[i]-left[i] >= 0)
		{


			_fmemset(p0+left[i],pen,right[i]-left[i]+1);

		}
		p0+=320;
	}
}

//rectangle(...) draws a rectangle, parallel with the screen edges
void TScreen::rectangle(int x1, int y1, int x2, int y2)
{
	if (x1<0) x1=0;
	if (y1<0) y1=0;

	if (x2>=width) x2=width-1;
	if (y2>=height) y2=height-1;

	if (x2<x1) return;
	if (y2<y1) return;


	for (int i=y1; i<=y2; ++i)
		_fmemset (buff[i+top]+x1+left, pen, x2-x1+1);
}

/********************* TAnaglyphScreen ************************/
/*
TAnaglyphScreen supports the anaglyph 3d-viewing technique.
The main difference from TScreen is that it has two
off-screen buffers, one for the left eye and one for the right eye.
You can switch between the eyes using selectEye(eye).

Draw the scene twice, once for each eye, shifting the camera between the
two shots. redraw() then creates an anaglyph image.
*/
TAnaglyphScreen::TAnaglyphScreen():TScreen()
{
	if (ok)
	{
		buffLeft=buff;	//use the inherited buffer for the left eye
		(void far*)buffRight=farmalloc(64000);	//get one for the right eye
		ok=buffRight?1:0;
	}
}

TAnaglyphScreen::~TAnaglyphScreen()
{
	if (buffRight) farfree(buffRight);
	buff=buffLeft;
	TScreen::~TScreen();
}

//selects the appropriate off-screen buffer
void TAnaglyphScreen::selectEye(char n)
{
	buff=n?buffRight:buffLeft;
}

//create an anaglyph image (from buffLeft and buffRight) on the screen
void TAnaglyphScreen::redraw()
{
	wfvr();

	unsigned int scrlo, scrhi, leftlo, lefthi, rightlo, righthi;

	scrlo=(unsigned long) scr;
	scrhi=(((unsigned long) scr)>>16)+(scrlo>>4);
	scrlo&=0xf;

	leftlo=(unsigned long) buffLeft;
	lefthi=(((unsigned long) buffLeft)>>16)+(leftlo>>4);
	leftlo&=0xf;

	rightlo=(unsigned long) buffRight;
	righthi=(((unsigned long) buffRight)>>16)+(rightlo>>4);
	rightlo&=0xf;


	asm	{
	cld

	push ds		//backup ds

		//we want: ES:DI = scr; DS:SI = buffLeft; SS:SI = buffRight;
		//!!!!!!! This assumes that the normalized offsets of
		//buffRight and buffLeft equal!!!!
	mov di, scrlo
	mov si, leftlo

	mov cx,32000	//32000 words

	mov es, scrhi

	mov bx, righthi	//we won't have access to local vars after "mov ds, mov ss..."

	mov ax, lefthi	//mov ds, lefthi

	push ax
	pop ds

	mov dx,ss		//backup ss
	mov ss, bx		//mov ss, righthi (not accessible now!)
	cli				//disable interrupts, since ss is modified
	}

looplbl:			// for (;cx>0;--cx) {

	asm{

	mov bx, ss:[si] //get BX from buffRight
	lodsw			//get AX from buffLeft, ++SI	//!!!dela chyby!!

	and bx, 0x0303
	and ax, 0x3c3c
	or ax,bx

	stosw			// store AX word in the video RAM; ++DI
	loop looplbl	// } //end for

	mov ss, dx		//restore the original ss
	pop ds			//...and ds
	sti				//re-enable interrupts
	}
}

