#include "sw.h"
#include "ship.h"
#include "space.h"
#include <math.h>
#include <stdlib.h>
#include "3d.h"
#include "const.h"
#include <alloc.h>
#include <stdio.h>
#define sgn(x) ((x)==0?0:( (x)>0? 1:-1 ))
#define sqr(x) ((x)*(x))

TFlyingObject::TFlyingObject(TObject *o, TSpace* spc, int _owner): TObject(o)
{
	owner=_owner;
	v=0;	//stand-still
	a.x=a.y=a.z=0;
	pos.x=pos.y=pos.z=0;	//will be set to something else anyway
	TQuat q={0,0,0,1};		//just a legal rotation to start with (quats can't be just random!!)
	quat=q;
	space=spc;

}

TFlyingObject::~TFlyingObject()
{
}

void TFlyingObject::move(float da_up, float da_left, float da_forward, float dt)
{
	//the acceleration in every direction should be between -1 and 1
	da_up*=SIDES_RATIO;
	da_left*=SIDES_RATIO;
	da_forward*=FWD_RATIO;


	a.x*=REDUCE;a.y*=REDUCE;a.z*=REDUCE;	//let the jets cool down :-)
	//change the acceleration vector
	a.x+=da_left;
	a.y+=da_up;
	a.z+=da_forward;

	//compute new velocity, rotated so that the current velocity points to (0,0,1)

	//calculate the new velocity....
	v=v*v+a.z*dt;		//E=(1/2)*m*v^2 !!!
	if (v<0.01) v=0.01;
	v=sqrt(v);
	if (v>maxspeed) v=maxspeed;

	TVector nv={0,0,-v};
	nv.x+=a.x*v*dt; nv.y+=a.y*v*dt;	//scale the acceleration proportionally to v

	float nv_size=sqrt(nv.x*nv.x + nv.y*nv.y + nv.z*nv.z);


	TQuat oldq=quat;

	if (fabs(nv.x)>0.00001 || fabs(nv.y)>0.00001)
	{
		//calculate the new rotation
		nv.x*=v/nv_size; nv.y*=v/nv_size; nv.z*=v/nv_size;

		TVector f={0,0,-1};

		TVector r;
		r.x = (nv.y * f.z) - (f.y * nv.z);
		r.y = (nv.z * f.x) - (f.z * nv.x);
		r.z = (nv.x * f.y) - (f.x * nv.y);

		float angle=acos(((f.x*nv.x + f.y*nv.y + f.z*nv.z)/v));

		TQuat q;
		q.w=cos(angle/2);

		float S=sin(angle/2)/sqrt(r.x*r.x + r.y*r.y + r.z*r.z);
		q.x=r.x*S;
		q.y=r.y*S;
		q.z=r.z*S;


		quatMult(&quat, &oldq, &q);
	} else
	{
		nv.x=nv.y=0; nv.z=-v;
	}

	TVector vv=getVVect();
	pos.x+=vv.x*dt; pos.y+=vv.y*dt; pos.z+=vv.z*dt;
}

TVector TFlyingObject::getVVect()
{
	//get the current speed and direction
	TVector r={0,0,-v};
	rotateVectors(&r, &quat);
	return r;
}

void TFlyingObject::explode(int size)	//create about size TExplosions
{
	for (int i=0; i<2*size; ++i)
	{
		space->add(new TExplosion(pos, space, size));
	}
}

char* TFlyingObject::dump(char* b)
{
	b = TObject::dump(b);
	*b++ = owner;
	return b;
}

char* TFlyingObject::read(char *b)
{
	b = TObject::read(b);	
	owner = *b++;
	return b;
	
}

/******************* TSHIP *****************************************/

TShip::TShip(TObject* o, TSpace* s, int ow):TFlyingObject(o, s, ow)
{
	maxspeed=MAX_SPEED;
	lockedID=0; controller=0;
	lives=MAX_LIVES;
	lastShot=lastMissile=0;
	shells=MAX_SHELLS;
	missiles=MAX_MISSILES;
	lastdist=0;
	lockedOwner = lockedLives = -1;
}



void TShip::lockedTick(TFlyingObject* locked, float dt)
//	follow a given object. If it's a ship, try to destroy it, if not (it's a goodie!) eat it.
{
	TVector lp=locked->getPos();
	float dist=sqrt(sqr(lp.x-pos.x)+sqr(lp.y-pos.y)+sqr(lp.z-pos.z))+1;

	lp.x-=pos.x; lp.y-=pos.y; lp.z-=pos.z;	//lp is the vector to the object
	TQuat q=quat;
	q.w=-q.w;
	rotateVectors(&lp, &q);	//rotate lp so that -lp.z points forward

	int accel;
	if (locked->type() == TYPE_SHIP)	//a ship?
	{
		float a=(sqr(lp.x)+sqr(lp.y))/sqr(lp.z);	//what is the angle betweeb my direction and lp?
		if (a<0.00001) cannon();				//small? fire the cannon!
		else if (a<0.01 && swrandom(1000)<5)		//larger? maybe try a missile?
			missile();

		accel=-1;							//default is slow-down
		if (lp.z<0 && dist>1000) accel=1;	//if the object is ahead, accelerate!
		if (lastdist/dist>1.005) accel=-1;	//if it's approaching too fast, slow down (dont' hit it, stay behind)
	} else		//a goodie
	{
		accel=lp.z<0?1:-1;	//accelerate if it's ahead, accelerate, else slow-down
	}
	move(-sgn(lp.y), -sgn(lp.x), accel, dt);	//change the direction towards the locked object
	lastdist=dist;	//save the current distance
}


char TShip::tick(float dt)
//this defines the ship's behaviour
{

	lastdt = dt;		//used in cannon();
	
	lastShot+=dt;		//increase time-since-shot/missile
	lastMissile+=dt;

	if (controller)		//human controlled?
	{
		move(controller->up, controller->left, controller->forward, dt);
			//move as told to

		//find the ship that is most straight-ahead and lock it if possible
		long l=0, minl;
		float minf=1e20;	//very big

		TFlyingObject* ship;
		do	//for all the objects in space
		{
			l=space->getNext(l);
			if (l && (ship=space->get(l))->type() == TYPE_SHIP)	//a ship?
			{
				TVector p=ship->getPos();
				p.x-=pos.x; p.y-=pos.y;  p.z-=pos.z;
				TQuat q=quat;
				q.w=-q.w;
				rotateVectors(&p, &q);
				if (p.z < -1000)		//ahead?
				{
					float a=(sqr(p.x)+sqr(p.y))/sqr(p.z);	//direction deeviation
					if (a<minf)		//best so far?
					{
						minf=a;
						minl=l;
					}
				}
			}
		} while (l);	//while any objects left

		if (minf>0.01) minl=0;	//not straigh enough to be locked
		lockedID=minl;
		controller->locked=minl;	//tell the controller (used when drawing the cockpit)
		if (minl) {
			lockedLives = space->get(minl)->getLives();
			lockedOwner = space->get(minl)->getOwner();
		} else lockedOwner=-1;
		if (controller->flags.missile) missile();	//launch a missile/fire if the pilet wishes to
		if (controller->flags.cannon) cannon();
	}
	else	//computer controlled
	{
//		/***********/return 1;
		if (space->get(lockedID))   //something locked?
		{
			TFlyingObject* o=space->get(lockedID);
			lockedTick(o, dt);	//chase it!
			if (swrandom(10000) < o->getLives()*10/MAX_LIVES) lockedID=0;
			//maybe try something else? But keep it locked if it's a half dead ship or a goodie (lives==1)!
		} else
		{   	//nothing locked
			//go for a walk :-)
			move(swrandom(3)-1, swrandom(3)-1, v<MAX_SPEED/2?1:swrandom(3)-1, dt);

			//see if we're not short of anything (shields, shells,...).
			//If so, look for the appropriate goodie
			long newL=0;
			if (swrandom(1000) < 20) newL=findRandomObject(TYPE_SHIP);
			if (newL) lockedID=newL;	//maybe lock someone?
			if (missiles<MAX_MISSILES/6) newL=findRandomObject(TYPE_MISSILES);
			if (newL) lockedID=newL;	//too few missiles?
			if (shells<MAX_SHELLS/8) newL=findRandomObject(TYPE_SHELLS);
			if (newL) lockedID=newL;	//too few shells? more important than missiles!
			if (lives<MAX_LIVES/2) newL=findRandomObject(TYPE_SHIELDS);
			if (newL) lockedID=newL;	//too few shields? more important than anything!

		}
	}
	return 0;
}


long TShip::findRandomObject(int type)
//find a random object of the given type in space
{
	int rnd=swrandom(50);		//the random factor
	long l=0, last=0;
	TFlyingObject* p;
	do	//Take rnd into account. Don't choose myself!
	{
		do
		{
			l=space->getNext(l);	//next object
			p=space->get(l);
			if (!p && last==0) return 0;   	//no such object in space
		} while (!p || p->type() != type);
		if (last==l && p==this) return 0;	//we're the only object in space
		last=l;
	}
	while(p==this || --rnd>0);		//repeat rnd times and make sure p is not this
	return l;
}

void TShip::collide(TFlyingObject* o)
//called when something collided into us
{
	lives-=o->collisionPower();		//decrease lives (shields)
	if (lives<=0)	//dead?
	{
		explode(50);		//explode
		die();				//die properly
	}

	switch (o->type())	//who did we collide into?
	{
		case TYPE_SHIP:		//a ship? move a bit, so that we don't stay inside each other (ugly+decreases shields)
			pos.x+=swrandom((long)(radius*10))-radius*5;
			pos.y+=swrandom((long)(radius*10))-radius*5;
			pos.z+=swrandom((long)(radius*10))-radius*5;
			break;
		case TYPE_SHIELDS:	//any of the goodies?
			lives=MAX_LIVES;
			break;
		case TYPE_SHELLS:
			shells+=200;
			if (shells>MAX_SHELLS) shells=MAX_SHELLS;
			break;
		case TYPE_MISSILES:
			missiles+=3;
			if (missiles>MAX_MISSILES) missiles=MAX_MISSILES;

			break;
	}

}

void TShip::missile()
//fire a missile, if possible
{
	if (missiles> 0 && lockedID && lastMissile > MISSILE_INTERVAL)
	//if any missiles left, and someone locked, and no missile launched recently
	{
		--missiles;		//used up one
		lastMissile=0;	//we launched now
		TMissile* m=new TMissile(space, owner, lockedID);
		m->setQuat(&quat);	//same direction as us

		TVector p={0,0,-2*(radius+m->getRadius())};		//launch in front of us
		rotateVectors(&p, &quat);
		p.x+=pos.x;	p.y+=pos.y; p.z+=pos.z;
		m->setPos(p);
		m->setV(v+MAX_SPEED/10);			//the same speed as us
		space->add(m);		//add into space
	}
}



void TShip::cannon()
{
	if (shells>0 && lastShot > CANNON_INTERVAL)
	//if any shells left and not fired recently
	{
		--shells;	//one shell used up
		lastShot=0;

		//the shells are so fast that they could "fly through" a ship without
		//the space being able to detect collison
		//we can solve that by firing more shots at once, with the distance between
		//any two smaller than the distance required to detect collision

		float speed=v+SHOT_SPEED;	//what will the speed be?
		float i=0;
		do
		{
			TShot* s=new TShot(space, owner);	//new shot
			s->setQuat(&quat);					//our direction
			float rad=s->getRadius()+radius;    //the distance between two shells
			TVector p={0,0,lastdt*SHOT_SPEED*0.8-i-rad-20};	//place the shells in front of us, in increasing distance
			rotateVectors(&p, &quat);
			p.x+=pos.x;	p.y+=pos.y; p.z+=pos.z;
			s->setPos(p);
			s->setV(speed);		//set the shot's speed
			space->add(s);		//add into space
			i+=2*rad;			//we covered the length of i
		} while (i<speed * lastdt);
		//enough? This is a SILLY hack. The dt from the last tick
		//is used to calculate the distance a shot will travel in
		//one tick.... couldn't think of a better way :-(
	}
}

char* TShip::dump(char* b)
{
	b = TFlyingObject::dump(b);
	int* bi = (int*) b;
	*bi++ = PACK(v, 1);
	*bi++ = PACK(lockedOwner, 1);
	*bi++ = PACK(lockedLives, 1);
	*bi++ = PACK(lives, 1);
	*bi++ = PACK(shells, 1);
	*bi++ = PACK(missiles, 1);
	return (char*) bi;
}

char* TShip::read(char *b)
{
	b = TFlyingObject::read(b);	
	int* bi = (int*) b;
	v = (int) UNPACK(*bi++, 1);
	lockedOwner = (int) UNPACK(*bi++, 1);
	lockedLives = (int) UNPACK(*bi++, 1);
	lives = (int) UNPACK(*bi++, 1);
	shells = (int) UNPACK(*bi++, 1);
	missiles = (int) UNPACK(*bi++, 1);
	return (char*) bi;
}



/******************* TMISSILE *****************************************/
//most methods similar to those of a TShip

TMissile::TMissile(TSpace* s, int ow, long lock):TFlyingObject(MISSILE ,s, ow)
{
	maxspeed=1.5*MAX_SPEED;	//1.5*faster than ships
	lockedID=lock;
	lives=1;
	lastdist=0;
	ttl = MISSILE_TTL;
}

void TMissile::collide(TFlyingObject* o)
{
	lives-=o->collisionPower();
	if (lives<=0)
	{
		explode(10);
		die();
	}

}

char TMissile::tick(float dt)
{

	TFlyingObject* l=space->get(lockedID);
	if (l)	//is the locked object still valid?
		lockedTick(l, dt);
	else die();	//no!

	if (--ttl < 0) die();
	return 0;
}


void TMissile::lockedTick(TFlyingObject* locked, float dt)
{
//	locked=space->get(space->getNext(space->getNext(0)));
	TVector lp=locked->getPos();
	float dist=sqrt(sqr(lp.x-pos.x)+sqr(lp.y-pos.y)+sqr(lp.z-pos.z))+1;

	lp.x-=pos.x; lp.y-=pos.y; lp.z-=pos.z;
	TQuat q=quat;
	q.w=-q.w;


	rotateVectors(&lp, &q);

	float accel=-1;
	if ((lastdist/dist>(1+.2/dt) || dist < getV() * dt *5) 
			&& getV() > 1.4*locked->getV() && getV() > MAX_SPEED/7)		//approaching fast
		accel=-1;			//slow down
	else if (lp.z<0 && -lp.z*1.2 > dist) accel=.7;	//otherwise accelerate

	move(-sgn(lp.y), -sgn(lp.x), accel, dt);
	lastdist=dist;
}



/******************* TSHOT *****************************************/
TShot::TShot(TSpace* s, int ow):TFlyingObject(SHELL ,s, ow)
{
	lives=1;
	ttl=SHOT_TTL;	//time-to-live
	radius*=SHOT_RADIUS_RATIO;	//pretend we're this big! improves hit probability -> makes the game more fun
}

void TShot::collide(TFlyingObject* o)
{
	lives-=o->collisionPower();
	if (lives<=0)
	{
		explode(8);
		die();
	}
}


char TShot::tick(float dt)
{
	ttl-=dt;	//decrease ttl
	if (ttl<0) die();	//too old?
	TVector w={0,0,-v*dt};	//fly straight
	rotateVectors(&w, &quat);
	pos.x+=w.x;
	pos.y+=w.y;
	pos.z+=w.z;
	return 0;
}


/******************* TExplosion *****************************************/
//This is an object representing a piece of garbage from an object that exploded
//the template is chosen randomly
TExplosion::TExplosion(TVector _pos, TSpace* s, int _size) : TFlyingObject(EXPLOSIONS[swrandom(EXPLOSIONSCNT)] ,s, 0)
{
	pos=_pos;
	size=_size+swrandom(_size);	//each piece has a different TTL
	V.x=swrandom(200)-100;		//random speed and direction
	V.y=swrandom(200)-100;
	V.z=swrandom(200)-100;
	//random axis of rotation
	TVector rot={swrandom(100)-50, swrandom(100)-50, swrandom(100)-50};
	float rlen=sqrt(sqr(rot.x)+ sqr(rot.y)+sqr(rot.z));
	float alpha=swrandom(10)/2/PI/2;	//random angle of rotation
	//make a quat from the axis and the angle
	Q.x=rot.x/rlen*sin(alpha);
	Q.y=rot.y/rlen*sin(alpha);
	Q.z=rot.z/rlen*sin(alpha);
	Q.w=cos(alpha);
}

void TExplosion::collide(TFlyingObject*)
{
}


char TExplosion::tick(float)
{
	if (--size <0) die();	//too old?
	TQuat q=quat;
	quatMult(&quat,&q, &Q);	//rotate
	pos.x+=V.x;	//fly straight
	pos.y+=V.y;
	pos.z+=V.z;
	return 0;
}

/******************* GOODIES *****************************************/
//These objects do nothing

TShells::TShells(TVector _pos, TSpace* s) : TFlyingObject(SHELLS ,s, 0)
{
	pos=_pos;
	lives=1;
}

TShields::TShields(TVector _pos, TSpace* s) : TFlyingObject(SHIELDS ,s, 0)
{
	pos=_pos;
	lives=1;
}

TMissiles::TMissiles(TVector _pos, TSpace* s) : TFlyingObject(MISSILES ,s, 0)
{
	pos=_pos;
	lives=1;
}



