#include "sw.h"
#include "sw_c.h"
#include "space.h"
#include "const.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <values.h>

#define ind(l) ((l)>>16)		//ID->index
#define chk(l) ((l)&0xffff)		//ID->chk number
#define sqr(x) ((x)*(x))

TVector nil={0,0,0};

TSpace::TSpace(int max, score_table* _scores)
{
	scores=_scores;
	ID=1+swrandom(30000);		//random ID
	maxObj=max;
	obj=new TFORef[max+2];
	sortedIndices=new TSortTag[max+1];

	for (int i=0; i<=maxObj; ++i) obj[i].o=0;	//all references are empty
}

TSpace::~TSpace()
{
	for (int i=0; i<maxObj; ++i) if (obj[i].o) delete obj[i].o;
	delete obj;
	delete sortedIndices;
}

TFlyingObject* TSpace::get(unsigned long ID)
{
	if (ind(ID)>=(unsigned int)maxObj || obj[ind(ID)].chk != chk(ID)) return 0;	//Is ID valid? NO!
	else return obj[ind(ID)].o;			//Yes
}


long TSpace::add(TFlyingObject* o)
{
	int i;
	++ID;		//new chk part of ID
	if (ID==0) ++ID;
	ID &= 0xFFFF;

	for (i=0; (obj[i].o) && (i<maxObj-1); ++i);	//get the first empty refernce

	obj[i].o=o;
	obj[i].chk=ID;

	return (((long)i)<<16) + obj[i].chk;	//return the ID
}

unsigned long TSpace::getNext(unsigned long prev)
{
	int i=-1;	//start from the first object
	if (obj[ind(prev)].chk==chk(prev)) i=ind(prev);	//is prev a valid ID? If yes, start there

	while (!obj[++i].o && i<=maxObj);	//get the first non-empty reference

	if (i>=maxObj) return 0;	//no more objects

	return (((long)i)<<16) + obj[i].chk;	//return the ID
}

void TSpace::del(unsigned long ID)
{
	if (get(ID))            	//a valid ID?
	{
		delete get(ID);			//delete the object
		obj[ind(ID)].o=0;		//mark the reference as empty
	}
}

void TSpace::draw(TScreen* scr, TCam* cam, TVector* light)
{
	distSort(cam->pos);		//sort the objects, so that the nearest get drawn first
	for (int i=sortedCnt-1; i>=0; --i)
		obj[ sortedIndices[i].index ].o->draw(scr, cam, light);	//draw all the objects
}

/************** drawing the map ************/
struct TPoint	//an on-screen point
{
	int x;
	int y;
};

//a function that maps a vector in the space to the screen position
TPoint map(TVector* v, TCam* cam, float zoom, int w, int h)
{
	TPoint r;
	float tmp=(cam->dist- (v->z - cam->pos.z))/cam->dist/zoom;

	r.x=(int)( (v->x-cam->pos.x-cam->shiftRight)/tmp + w/2 + cam->shiftRight*zoom );
	r.y=(int)( (v->y-cam->pos.y)/tmp );

	return r;
}


void TSpace::drawMap(TScreen* scr, TCam* cam, char local)
//draws the map
//if local is nonzero, only objects that are up to a constant distance away are drawn
//otherwise the map is scaled to fit in all the objects
{

	float mapgrid=MAPGRID;

	float distance=20000;

	if (!local)
	{			//scale the map
		distance=sqr(distance);
		//find the furthest object
		for (int i=0; i<maxObj; ++i)
			if (obj[i].o && obj[i].o->type() != TYPE_SHOT && obj[i].o->type() != TYPE_EXPLOSION)
			{
				TVector v=obj[i].o->getPos();
				v.x-=cam->pos.x;
				v.y-=cam->pos.y;
				v.z-=cam->pos.z;
				float tmp=sqr(v.x)+sqr(v.y)+sqr(v.z);

				if ( tmp > distance) distance=tmp;
			}
		distance=sqrt(distance);	//distance is the distance of the furthest object
	}

	int w=scr->getwidth();
	int h=scr->getheight();


	TCam c;
	c.pos.x=0;
	c.pos.y=-distance;      //move back so that all the map can be seen
	c.pos.z=0;
	c.dist=5*distance;		//a nice eye distance
	c.shiftRight=cam->shiftRight*c.dist/cam->dist;

	int linecnt=(int)(distance/mapgrid);	//how many grids?
	float zoom=1/(2.4*distance/(w>h?h:w));	//zoom to fit everything on the screen


	//first, draw the grid
	scr->setpen(shade(clWhite, 128));	//gray

	int i;
	for (i=-linecnt; i<=linecnt; ++i)	//"horizontal"
	{
		TVector a={-linecnt*mapgrid,0,i*mapgrid};
		TVector b={+linecnt*mapgrid,0,i*mapgrid};

		TPoint ma=map(&a, &c, zoom, w,h);
		TPoint mb=map(&b, &c, zoom, w,h);

		scr->line(ma.x, ma.y, mb.x, mb.y);
	}

	for (i=-linecnt; i<=linecnt; ++i)		//"vertical"
	{
		TVector a={i*mapgrid,0,-linecnt*mapgrid};
		TVector b={i*mapgrid,0,+linecnt*mapgrid};

		TPoint ma=map(&a, &c, zoom, w,h);
		TPoint mb=map(&b, &c, zoom, w,h);


		scr->line(ma.x, ma.y, mb.x, mb.y);
	}


	//draw the objects


	for (i=0; i<maxObj; ++i)
	{
		TFlyingObject* o=obj[i].o;
		if (o && o->type() != TYPE_SHOT && o->type() != TYPE_EXPLOSION)
			//don't draw shots and explosions
		{

			TVector t=o->getPos();
			scr->setpen(o->getAvgColour());

			t.x-=cam->pos.x; t.y-=cam->pos.y;t.z-=cam->pos.z;
			float dist=sqr(t.x)+sqr(t.y)+sqr(t.z);
			if (dist<=sqr(distance) && dist > sqr(200.0))
				//objects in not too far and not too near (the ship with the camera)
			{

				TQuat q=cam->quat; q.w=-q.w;
				rotateVectors(&t, &q);

				t.y=-t.y;
				TPoint a=map(&t, &c, zoom, w,h);	//point of the arrow

				char up=t.y>0;
				t.y=0;
				TPoint b=map(&t, &c, zoom, w,h);	//end of the arrow (on the grid plane)

				scr->line(a.x,a.y, b.x, b.y);		//draw the vertical line
				//draw a little arrow/cross;
				scr->putPixel(a.x, a.y, -1);
				if (up) --a.y; else ++a.y;
				scr->putPixel(a.x-1, a.y, -1);
				scr->putPixel(a.x, a.y, -1);
				scr->putPixel(a.x+1, a.y, -1);

			}
		}
	}



}


void TSpace::tick(float dt)
//tick all the objects, detect collisions, remove dead objects
{
	int i;
	float maxRadius=0;
	// a two purpose cycle
		//get the maximum radius
		//tick all the objects
	for (i=0; i<maxObj; ++i)
	{
		if (obj[i].o)	//not empty?
		{
			obj[i].o->tick(dt);		//tick
			if (obj[i].o->getRadius() > maxRadius) maxRadius=obj[i].o->getRadius();	//maximum?
		}
	}

	//sort the objects according to the distance from a certain point (why not 0...)
	distSort(nil);

	sortedIndices[sortedCnt].x=MAXFLOAT;

	for(i=0; i<sortedCnt-1; ++i)	//for all the objects
	{
		//'current' is the i'th object
		TFlyingObject* current=obj[sortedIndices[i].index].o;
		float cd=sortedIndices[i].x;	//current's distance
		float cr=current->getRadius();	//current's radius
		TVector cp=current->getPos();	//current's position


		//let's say 'current' collides with an object o.
		//if o's distance is < current's, the collision has already been detected.
		//o's index must be > i and o's distance must be < current's distance + current's radius + o's radius
		//o's diameter is <= maxRadius

		int j=i+1;	//start with i+1
		while(sortedIndices[j].x < cd+cr+maxRadius)	//see above
		{
			TFlyingObject* other=obj[sortedIndices[j].index].o;
			TVector op=other->getPos();
			float _or=other->getRadius();
			if (sqr(cp.x-op.x) + sqr(cp.y-op.y) + sqr(cp.z-op.z) < sqr(_or+cr))	//do they really collide?
			{
				//call the both object's collide(...) methods,
				//	only if they are still alive (otherwise a chain reaction could occur with explosions for example)
				if (current->getLives() >0 ) current->collide(other);
				if (other->getLives() >0 )	other->collide(current);

				//update the scoretable
				if (current->type() == TYPE_SHIP && current->getLives()<0)
					//current has been killed by other
				{
					scores[current->getOwner()].deaths++;
					scores[other->getOwner()].kills++;

				}

				if (other->type() == TYPE_SHIP && other->getLives()<0)
					//other has been killed by current
				{
					scores[current->getOwner()].kills++;
					scores[other->getOwner()].deaths++;
					
				}

			}
			++j;
		}
	}
	removeDead();	//remove dead objects

}

//used for quicksort
extern "C" int sortTagCmp(const void* a, const void* b)
{
	return ( ((TSortTag*) a)->x > ((TSortTag*) b)->x )*2-1;
}

void TSpace::distSort(TVector v)
	//sort the objects according to their distance from v
{
	sortedCnt=0;
	//fill the sortedIndices array with only the non-empty references.
	for (int i=0; i<maxObj; ++i)
		if (obj[i].o)
		{
			sortedIndices[sortedCnt].index=i;
			TVector op=obj[i].o->getPos();
			sortedIndices[sortedCnt].x=sqrt(sqr(v.x-op.x)+sqr(v.y-op.y)+sqr(v.z-op.z));
			//calculate the distance
			++sortedCnt;
		}

	qsort(sortedIndices, sortedCnt, sizeof (*sortedIndices), sortTagCmp);
		//sort
}

void TSpace::removeDead()
//remove objects whose lives <0
{
	for (int i=0; i<maxObj; ++i)
		if (obj[i].o && obj[i].o->getLives()<=0)
		{
			delete obj[i].o;
			obj[i].o=0;
		}
}

char *TSpace::dump(char *buff)
{
	for (int i=0; i<maxObj; ++i)
		if (obj[i].o) {
			*buff++ = obj[i].o->type();
			buff = obj[i].o->dump(buff);
		}
	*buff++ = TYPE_NOTHING;
	return buff;
}

void *TSpace::read(char *buff)
{
	TFlyingObject *o;
	while (*buff != TYPE_NOTHING) {
		switch (*buff) {
			case TYPE_SHIP: 
				o=new TShip(SHIP, this, 0); 
				break;
			case TYPE_MISSILE: 
				o=new TMissile(this, 0, 0);
				break;
			case TYPE_EXPLOSION:
				o=new TExplosion(nil, this, 1);
				break;
			case TYPE_SHOT:
				o=new TShot(this, 0);
				break;
			case TYPE_SHELLS:
				o=new TShells(nil, this);
				break;
			case TYPE_SHIELDS:
				o=new TShields(nil, this);
				break;
			case TYPE_MISSILES:
				o=new TMissiles(nil, this);
				break;
			default:
				printf("ooops! strange object type!\n");
		}
		
		//read the quat and position
		buff = o->read(buff+1);
		add(o);
	}	
	return buff+1;
}

void TSpace::delAll()
//remove all objects 
{
	for (int i=0; i<maxObj; ++i)
		if (obj[i].o)
		{
			delete obj[i].o;
			obj[i].o=0;
		}
}


