#ifndef __SPACE
#define __SPACE
class TSpace;
struct TFORef;

#include "ship.h"
#include "3d.h"
#include "screen.h"
#include "sw.h"

struct TFORef		//an object's reference
{
	unsigned int chk;
	TFlyingObject* o;
};

struct TSortTag		//used for sorting objects according to a distance
{
	int index;
	float x;
};

//this is the class representing the game's space
//it holds objects, moves them, detects collisions, removes dead objects,
//draws the screen, draws the map....

//the objects are kept in an array of TFORef structures
//each object is assigned a "chk" number which, together with its index, forms
//the objects ID. Pointers to objects obtained through get(...) must not be kept
//longer than till the next call of any TSpace method
//instead the IDs should be kept and pointers retrieved with get(...).


class TSpace
{
public:
	TSpace(int max, score_table *_scores);	//maximum objects
	~TSpace();

	TFlyingObject* get(unsigned long ID);	//get a pointer of the object, DON'T KEEP IT
	unsigned long getNext(unsigned long prev);		//if prev is a valid ID, returns the ID of the next object (0 after the last one)
									//if prev is not a valid ID, the first object is returned
	long add(TFlyingObject* o);		//add to space, the ID is returned

	void del(unsigned long ID);				//remove from space

	void draw(TScreen* scr, TCam* cam, TVector* light);	//draw the space
	void drawMap(TScreen* scr, TCam* cam, char local);	//draw a map
	void tick(float dt);	//tick all the objects, detect collisions, remove dead objects
	void delAll();
	char* dump(char* buff);
	void* read(char* buff);
private:

	TFORef* obj;				//the object references
	TSortTag* sortedIndices;	//an array of sorted indices to the obj array
	int sortedCnt;				//number of sorted objects
	void distSort(TVector);		//sort the objects by the distance from the given vector
	void removeDead();
	score_table* scores;				//the game's scoretable

	int _objCount;				//number of objects
	int maxObj;					//maximum number of objects
	long ID;					//las assigned ID

};


#endif

