#include "sw.h"
#include "screen.h"
#include "object.h"
#include "3d.h"
#include "ship.h"

#include <alloc.h>
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include "const.h"
#include "sw_c.h"

#define sqr(x) ((x)*(x))

// global variables (communication with GUI)
int _missiles, _shells, _shields, _speed, _my_number, _locked, _locked_lives;
// global object templates

TAnaglyphScreen *screen;	//the main graphical screen


/*********************************************************
 ************** MAIN
 *********************************************************/
char _map_local=1;	//1 if the map scale should stay constant, 0 for dynamic re-scaling
int _eye_dist=0;		//a default eye distance (can be changed during game)


//********** camera and light settings
TVector light={1,0,0};	//light direction
TCam c;

long myShip;

extern "C" char* sw_read(char* buff)
{
	space->delAll();
	return (char*)space->read(buff);
}

extern "C" void sw_client_init(void * window)
{
	screen = new TAnaglyphScreen(window);
}

//**********************
//**** DRAWING
//**********************
extern "C" int sw_draw()
{
	TShip *mine = NULL;
	unsigned long o=space->getNext(0);
	while (o && !mine) {
		TFlyingObject* f=space->get(o);
		if (f->type() == TYPE_SHIP && f->getOwner() == _my_number) {
			mine=(TShip*)f;
			
		}
		o = space->getNext(o);
		
	}
	screen->configure();
	int scrw=screen->getscrwidth();
	int scrh=screen->getscrheight();
	c.dist=1.1*scrw;	//eye-screen distance

	unsigned int background=shade(clRed, mine?25:150);
		//red if I'm dead, otherwise very-very-dark-red (better with anaglyph glasses on)

	//******** set the camera position, orientation, etc.
	if (mine)	//i'm alive
	{
		c.quat=mine->getQuat();	//same direction as my ship
		c.pos=mine->getPos();	//same position

		//move forward so my ship can not be seen

		TVector v={0,0,-c.dist};
		TQuat q=mine->getQuat();
		rotateVectors(&v, &q);
		c.pos.x+=v.x;
		c.pos.y+=v.y;
		c.pos.z+=v.z;
	}
       	
	screen->stereo = (_eye_dist != 0);
	//for both eyes
	for (int eye=0; eye<(screen->stereo?2:1); eye++)
	{
		screen->selectEye(eye);	//make it 3D ! :-)
		c.shiftRight=eye?-_eye_dist:+_eye_dist;
		c.shiftRight*=(float)scrw/320;
		screen->clipRect(0,0,scrw,scrh);	//main window
		screen->clrscr(background);

		int w=screen->getwidth();
		int h=screen->getheight();

		space->draw(screen, &c, &light);	//draw the space !

		//draw the box in the middle of the screen
		int len=w/15;
		unsigned long frame=/*ctrl.locked*/1?clYellow:clWhite;
			//yellow if a shiip is locked, otherwise white
		screen->setpen(frame);
		screen->rectangle(w/2-len, h/2-len, w/2+len, h/2-len);
		screen->rectangle(w/2-len, h/2+len, w/2+len, h/2+len);
		screen->rectangle(w/2+len, h/2-len, w/2+len, h/2+len);
		screen->rectangle(w/2-len, h/2-len, w/2-len, h/2+len);

		//draw the shield status indicator of the locked ship
		
		if (mine && mine->lockedOwner >= 0)
		{
			int x1=w/2-w/10;
			int x2=w/2+w/10;
			int y1=h/2-w/10;
			int y2=h/2+w/10;
			
			screen->line(x1, y1, x2, y1);
			screen->line(x1, y2, x2, y2);
			screen->line(x1, y1, x1, y2);
			screen->line(x2, y1, x2, y2);

			_locked = mine->lockedOwner;
			_locked_lives = mine->lockedLives;
			
		} else _locked = -1;
		screen->putPixel(w/2,h/2, frame);	//the dot in the centre

		//draw the map
		screen->clipRect(scrw*4/7+1,scrh*4/7+1,scrw-1,scrh-1);	//bottom right
		space->drawMap(screen, &c, _map_local);


		//***draw the indicators
		

	}	//for 2 eyes
	screen->redraw();

	if (mine) {
		_shields = (int)mine->getLives();
		_speed = (int)mine->getV();
		_missiles = (int)mine->getMissiles();
		_shells = (int)mine->getShells();
	}

	return 0;
}



