#include "sws.h"
#include "sw_c.h"
#include "const.h"

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>

char * ALIEN_NAMES[] = {"Osama", "Adolf", "Benito", "Jasir", "Bill", "Vladimir I.", "Josip V."};
#define ALIEN_NAME_COUNT  (sizeof(ALIEN_NAMES) / sizeof(char*))

struct client {
	int number;
	int socket;
	char keys;
	int answer;
	char message[100];
};

char global_message[200];



void disconnect(struct client *c, int *clientcnt, int n)
{
	close (c[n].socket);
	printf("Disconnect\n");
	sw_kill_client(c[n].number);
	strcat(global_message, _scores[c[n].number].name);
	strcat(global_message, " has left the game\n");
	*_scores[c[n].number].name = _scores[c[n].number].score =  _scores[c[n].number].deaths = _scores[c[n].number].kills =0;
	
	c[n]=c[*clientcnt - 1];
	--(*clientcnt);
	
}

int main()
{
	struct linger linger;
	struct client clients[MAX_CLIENTS];
	long aliens[MAX_SHIPS];
	
	int clientcnt=0, aliencnt=0;
	struct timeval tm, last_tm;
	float dt;
	struct score_table oldscores[MAX_SHIPS];
		
	int sock1;
	struct sockaddr_in my_addr;
	int i;

	/* fork, and make one of the processes send a broadcast datagram
	 * repeatedly */
	if (fork() == 0) {
		int on=1;
		sock1 = socket(AF_INET, SOCK_DGRAM, 17);
		if (sock1 == -1) {
			perror("creating UDP socket");
			return 1;
		}
		
		my_addr.sin_family = AF_INET;
		my_addr.sin_port = htons(SERVER_PORT);
		my_addr.sin_addr.s_addr = INADDR_BROADCAST;

		if (setsockopt(sock1, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on))==-1) {
			perror("setsockopt (allowing broadcasts)");
			return 1;
		}
		
		while (1) {
			sleep(2);
			sendto(sock1, "X", 1, 0, (struct sockaddr*)&my_addr, sizeof(my_addr));
			
		}
	}

	
	for (i=0; i< MAX_SHIPS; ++i) {
		*_scores[i].name = 0;
		_scores[i].kills = _scores[i].deaths = _scores[i].score =0;
	}
	
	sw_init(); 	//initialize the space
	printf("Space initialized\n");

	
	sock1 = socket(AF_INET, SOCK_STREAM, 6);
	if (sock1 == -1) {
		perror("creating socket");
		return 1;
	}

	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons(SERVER_PORT);
	my_addr.sin_addr.s_addr = INADDR_ANY;
	if ( bind(sock1, (struct sockaddr*)&my_addr, sizeof(my_addr)) == -1) {
		perror("bind");
		return 1;
	}

	listen(sock1, MAX_CLIENTS);

	/* shutdown the socket immediatelly on close */
	linger.l_onoff = 1;
	linger.l_linger = 0;
	if (setsockopt(sock1, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger)) == -1)
		perror("not critical, setsockopt");
	
	last_tm.tv_sec=0;

	while(1) {
		int i, maxfd;
		int answers = 0;
			
		for (i=0; i<MAX_CLIENTS; ++i) clients[i].answer=0;
		
		*global_message=0;	//clear the message
		while (answers < clientcnt || clientcnt==0) {
			fd_set fdset, ex_fdset;
			struct timeval timeout;
			int select_ret;
			timeout.tv_sec=10;
			timeout.tv_usec=0;
			
			FD_ZERO(&fdset);
			FD_ZERO(&ex_fdset);
			/* check for new clients */
			FD_SET(sock1, &fdset);

			/* check for data from clients */
				maxfd = sock1;
			for (i=0; i<clientcnt; ++i) {
				FD_SET(clients[i].socket, &fdset);	
				FD_SET(clients[i].socket, &ex_fdset);	
				if(clients[i].socket > maxfd) maxfd=clients[i].socket;
			}
			
			select_ret= select(maxfd+1, &fdset, NULL, &ex_fdset, &timeout);
			if (select_ret == -1) perror ("select");		


			/* new client connecting? */
			if (FD_ISSET(sock1, &fdset)) {
				int sock;
				struct sockaddr_in addr;
				socklen_t addrlen;

				/* accept the connection*/
				addrlen = sizeof(addr);
				sock = accept(sock1, (struct sockaddr*)&addr, &addrlen);
				if (sock == -1) perror("accept");
				else {

					printf("new client: %s\n", inet_ntoa(addr.sin_addr));
				
					/* out of capacity? */
					if (clientcnt >= MAX_CLIENTS) {
						close(sock);
						fprintf(stderr, "too many clients\n");
					} 
					/* no, add the client */
					else {
						
						/* new ship */
						clients[clientcnt].number = sw_new_client(0); 
						if (clients[clientcnt].number < 0) {
							fprintf(stderr, "too many ships\n");	
							close(sock);
						} 
						else {
							*_scores[clients[clientcnt].number].name = 0;
							clients[clientcnt++].socket = sock;
						
							/* non blocking io */
							fcntl(sock, F_SETFL, O_NONBLOCK);
						}
						
					}
				}
			}

			/* data from a client */
			for (i=0; i< clientcnt; ++i) {
				if (FD_ISSET(clients[i].socket, &ex_fdset))
					disconnect(clients, &clientcnt, i);
				else if (FD_ISSET(clients[i].socket, &fdset)) {
					char buffer[100];
					memset(buffer, 0, 100);
					if (read(clients[i].socket, buffer, 99)<=0)
						disconnect(clients, &clientcnt, i);
					else {
						clients[i].answer=1;
						++answers;
						clients[i].keys=buffer[0];
						strcpy(clients[i].message, buffer+1);
						if (!*_scores[clients[i].number].name) {
							strncpy(_scores[clients[i].number].name, clients[i].message, 15);
							_scores[clients[i].number].name[15]=0;
							strcat(clients[i].message, " has joined the game");
							*oldscores[clients[i].number].name = 0;
						}
						
							
						
					}
				}
			}
			
			/* timeout occured */
			if (select_ret==0) 
				for (i=0; i<clientcnt; ++i) 
					if (!clients[i].answer)
						disconnect(clients, &clientcnt, i);
	
		} /* while (answers < ..... */



		/* compute the time delta since last iteration and "tick" the space */
		gettimeofday(&tm, NULL);

		/* special handling of first iteration */
		if (last_tm.tv_sec != 0) 
			dt = (tm.tv_sec - last_tm.tv_sec) + (float)(tm.tv_usec - last_tm.tv_usec)/1E6;
		else dt=0.1;
		last_tm = tm;
		//printf("Framerate: %f \n ", 1/dt);

		for (i=0; i< clientcnt; ++i)
			sw_set_keys(clients[i].number, clients[i].keys);


		if (aliencnt > clientcnt * ALIENS_PER_CLIENT + 1) 
			if ((rand() & 63) == 1) {
				int i;
				i=rand()%aliencnt;
				sw_kill_client(aliens[i]);
				strcat(global_message, _scores[aliens[i]].name);
				strcat(global_message, " (alien) has left the game\n");
				*_scores[aliens[i]].name=0;
				
				aliens[i] = aliens[--aliencnt];	
			}
		       	

		if (aliencnt < clientcnt * ALIENS_PER_CLIENT)
			if ((rand() & 63) == 1) {
				
				aliens[aliencnt]=sw_new_client(1);
				strcpy(_scores[aliens[aliencnt]].name, ALIEN_NAMES[aliencnt%ALIEN_NAME_COUNT]);
				strcat(global_message, _scores[aliens[aliencnt]].name);
				strcat(global_message, " (alien) has joined the game\n");
				++aliencnt;
			}


		sw_tick(dt*28);	
		{
			char send_buff[4*8 * MAX_OBJECTS+100*MAX_CLIENTS+MAX_SHIPS*4*4];
			char* p;
			char message=0;

			/* dump the space */
			p = sw_dump(send_buff+5);
			
			
			/* has the score table changed? */
			if (memcmp(_scores, oldscores, sizeof(*oldscores)*MAX_SHIPS)) {
				memcpy(oldscores, _scores,  sizeof(*oldscores)*MAX_SHIPS);
				*p++=1;
				p = dump_score_table(p);
			} else *p++ = 0;
			
			
			*p=0;
			for (i=0; i<clientcnt; ++i) {
				if (*clients[i].message) {
				       	strcat(p, clients[i].message);
					strcat(p, "\n");
					message =1;
				}
			}
		
			for (i=0; i<_dead_count; ++i) {
				strcat(p, _scores[_dead[i]].name);
				strcat(p, " has been killed\n");
				message =1;
			}

			if (*global_message) {
				strcat(p, global_message);
				message=1;
			}
				
			if (message) {
				p += strlen(p);
				--p;
			}
			*p++=0;	
			
			*((int*)(send_buff+1)) = htons(p-send_buff);
				
						
			for (i=0; i<clientcnt; ++i) {
				*send_buff = clients[i].number;
				if (write(clients[i].socket, send_buff, p-send_buff) != p - send_buff)
					perror("write");
				
			}
			
			
			

		}			
//		usleep(100000);	
		



	} /* main loop */
		

}
