#include "sws.h"
#include "sw.h"

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include "const.h"

struct TClient {
	long ship;
	TController ctrl;
	char alien;
} clients[MAX_SHIPS];

int _dead[MAX_SHIPS];
int _dead_count;

void newGoodie(TSpace* space, int type)
//add a new shells/missiles/shields object to the space
{
	TVector p;
	p.x=swrandom(10000)-5000;
	p.y=swrandom(10000)-5000;
	p.z=swrandom(10000)-5000;

	TFlyingObject* o;
	switch(type)
	{
		case TYPE_SHIELDS:
			o=new TShields(p, space);
			break;
		case TYPE_MISSILES:
			o=new TMissiles(p, space);
			break;
		case TYPE_SHELLS:
			o=new TShells(p, space);
			break;
		default:
			return;
	}
	space->add(o);
}

long newShip(TSpace* space, TController* ctrl, int owner)
//add a new ship to the space, returning its ID
{
	TShip* s=new TShip(SHIP, space, owner);
	s->controller=ctrl;

	TVector p;
	p.x=swrandom(10000)-5000;
	p.y=swrandom(10000)-5000;
	p.z=swrandom(10000)-5000;

	s->setPos(p);

	return space->add(s);
}

extern "C" int  sw_new_client(char alien) 
{
	static int first_run=1;
	int i;

	if(first_run) {
		for (i=0; i<MAX_SHIPS; ++i)
			clients[i].ship=0;
		first_run=0;
	}

	i=-1;
	while (clients[++i].ship != 0);
		
	clients[i].alien=alien;
	clients[i].ship = 
		newShip(space, clients[i].alien?0:&(clients[i].ctrl), i);
	
	_scores[i].kills = _scores[i].deaths = _scores[i].score =0;
	
	return i;
}

extern "C" void sw_kill_client(int n) 
{
	TFlyingObject* o=space->get(clients[n].ship);
	if (o) o->die();
	clients[n].ship = 0;
}

extern "C" void sw_set_keys(int client, char keys) {
	char up=0, left=0, fwd=0;
	if (keys&KEY_LEFT) ++left;
	if (keys&KEY_RIGHT) --left;
	if (keys&KEY_UP) ++up;	
	if (keys&KEY_DOWN) --up;
	if (keys&KEY_FW) ++fwd; 
	if (keys&KEY_BACK) --fwd;
	clients[client].ctrl.flags.missile=(keys&KEY_MISSILE)!=0; 
	clients[client].ctrl.flags.cannon=(keys&KEY_CANNON)!=0;
	clients[client].ctrl.left=left; 
	clients[client].ctrl.up=up;
	clients[client].ctrl.forward=fwd;

}

extern "C" char* sw_dump(char* buff)
{
	return space->dump(buff);
}


extern "C" int sw_tick(float deltat)
{

//********* do everything :-)

	space->tick(deltat);

//********* check for missing ships and goodies and replace them

	int shellsCount=0, missilesCount=0, shieldsCount=0;
	long l=0;
	char shipsAlive[MAX_SHIPS];
	int i;
	for (i=1; i<MAX_SHIPS; ++i) shipsAlive[i]=0;

	//count ships and goodies
	while ((l=space->getNext(l)))
	{
		TFlyingObject* o=space->get(l);
		switch (o->type())
		{
			case TYPE_SHIP:
				shipsAlive[o->getOwner()]=1;	//this ship is alive
				break;
			case TYPE_MISSILES:
				++missilesCount;
				break;
			case TYPE_SHIELDS:
				++shieldsCount;
				break;
			case TYPE_SHELLS:
				++shellsCount;
				break;
		}
	}
	//replace missing goodies
	if (shieldsCount < 4 && swrandom(1000)<5 ) newGoodie(space, TYPE_SHIELDS);
	if (shellsCount < 1 && swrandom(1000)<5 ) newGoodie(space, TYPE_SHELLS);
	if (missilesCount < 2 && swrandom(1000)<5 ) newGoodie(space, TYPE_MISSILES);

	//fill missing ships, let clients return to the game after being
	//killed
	
	_dead_count=0;
	for (i=0; i<MAX_SHIPS; ++i)
		if (clients[i].ship) {
		        if(!space->get(clients[i].ship)) {
				if (clients[i].ship != -1) _dead[_dead_count++] = i;
				clients[i].ship = -1;	//don't report we're dead more than once

				if ((clients[i].alien && rand()%128 == 1) || (!clients[i].alien && clients[i].ctrl.flags.missile)) {
					//printf("------------ %d, %s, %d\n", i, _scores[i].name, clients[i].alien);
					clients[i].ship = newShip(space, clients[i].alien?0:&(clients[i].ctrl), i);
				}
				
			}
		}			

			
	return 0;
}

 
